@@ -3,22 +3,33 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 2013, VU University Amsterdam
+    Copyright (c)  2009-2017, VU University Amsterdam
+    All rights reserved.
 
-    This library is free software; you can redistribute it and/or
-    modify it under the terms of the GNU Lesser General Public
-    License as published by the Free Software Foundation; either
-    version 2.1 of the License, or (at your option) any later version.
+    Redistribution and use in source and binary forms, with or without
+    modification, are permitted provided that the following conditions
+    are met:
 
-    This library is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-    Lesser General Public License for more details.
+    1. Redistributions of source code must retain the above copyright
+       notice, this list of conditions and the following disclaimer.
 
-    You should have received a copy of the GNU Lesser General Public
-    License along with this library; if not, write to the Free Software
-    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
-    02110-1301  USA
+    2. Redistributions in binary form must reproduce the above copyright
+       notice, this list of conditions and the following disclaimer in
+       the documentation and/or other materials provided with the
+       distribution.
+
+    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
+    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
+    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
+    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
+    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
+    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
+    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
+    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
+    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
+    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
+    POSSIBILITY OF SUCH DAMAGE.
 */
 
 #include <config.h>
@@ -157,6 +168,12 @@ typedef enum format
   D_TRIG_NO_GRAPH			/* read TRiG as Turtle */
 } format;
 
+typedef enum recover_mode
+{ R_STATEMENT = 0,			/* on error, re-sync at statement */
+  R_PREDOBJLIST,			/* on error, re-sync at ; or . */
+  R_OBJLIST				/* on error, re-sync at , ; or . */
+} recover_mode;
+
 typedef struct resource
 { resource_type  type;
   int		 constant;		/* read-only constant */
@@ -220,6 +237,8 @@ typedef struct turtle_state
   resource     *free_resources;		/* Recycle resources */
   IOSTREAM     *input;			/* Our input */
   int		current_char;		/* Current character */
+  recover_mode	recover;		/* R_*: how to recover */
+  recover_mode  recovered;		/* R_*: how we recovered */
   error_mode	on_error;		/* E_* */
   format	format;			/* D_* */
   size_t	error_count;		/* Number of syntax errors */
@@ -259,6 +278,17 @@ static resource resource_constants[] =
   RES_CONST(XSD_BOOLEAN)	/* 7 */
 };
 
+#define RECOVER_BEGIN(mode) { recover_mode __or = ts->recover; \
+			      ts->recover = mode; \
+			      ts->recovered = R_STATEMENT;
+#define RECOVER_END()         ts->recover = __or; \
+			    }
+#define RECOVER(mode, code) \
+        RECOVER_BEGIN(mode); \
+	code; \
+	RECOVER_END();
+
+
 		 /*******************************
 		 *	       BUFFER		*
 		 *******************************/
@@ -417,13 +447,15 @@ is_scheme_char(int c)
 }
 
 static inline int
-is_lang_char1(int c)
-{ return (c < 128 ? (char_type[c] & (LC|UC)) != 0 : FALSE);
-}
+is_lang_char(int c, int minus)
+{ if ( c < 128 )
+  { if ( minus == 0 )			/* main language */
+      return (char_type[c] & (LC|UC));
+    else				/* sub language */
+      return (char_type[c] & (LC|UC|DI));
+  }
 
-static inline int
-is_lang_char(int c)
-{ return (c < 128 ? (char_type[c] & (LC|UC|DI)) != 0 : FALSE) || c == '-';
+  return FALSE;
 }
 
 static inline int
@@ -566,24 +598,7 @@ add_string_hash_map(hash_map *hm,
 static int	next(turtle_state *ts);
 
 static int
-print_warning(term_t t)
-{ static predicate_t print_message2;
-  term_t av;
-
-  if ( !print_message2 )
-    print_message2 = PL_predicate("print_message", 2, "system");
-
-  if ( (av = PL_new_term_refs(2)) &&
-       PL_put_atom(av+0, ATOM_warning) &&
-       PL_put_term(av+1, t) )
-    return PL_call_predicate(NULL, PL_Q_PASS_EXCEPTION, print_message2, av);
-
-  return FALSE;
-}
-
-
-static int
-syntax_message(turtle_state *ts, const char *msg, int is_error)
+print_message(turtle_state *ts, term_t ex1, int is_error)
 { term_t ex;
   IOPOS *pos;
 
@@ -591,12 +606,11 @@ syntax_message(turtle_state *ts, const char *msg, int is_error)
     return FALSE;
 
   ts->error_count++;
-
   if ( !(ex=PL_new_term_refs(2)) ||
-       !PL_unify_term(ex+0, PL_FUNCTOR, FUNCTOR_syntax_error1,
-		              PL_CHARS, msg) )
+       !PL_put_term(ex+0, ex1) )
     return FALSE;
 
+					/* create position term */
   if ( (pos=ts->input->position) )
   { term_t stream;
     int linepos = pos->linepos;
@@ -618,30 +632,68 @@ syntax_message(turtle_state *ts, const char *msg, int is_error)
       return FALSE;
   }
 
-  if ( PL_cons_functor_v(ex, FUNCTOR_error2, ex) )
-  { if ( is_error )
-    { for(;;)
-      { if ( !next(ts) || ts->current_char == -1 )
+					/* create error(Format, Pos) */
+  if ( !PL_cons_functor_v(ex, FUNCTOR_error2, ex) )
+    return FALSE;
+
+  if ( is_error )			/* re-sync after error */
+  { for(;;)
+    { if ( !next(ts) || ts->current_char == -1 )
+	break;
+      if ( ts->current_char == '.' )
+      { if ( !next(ts) ||
+	     ts->current_char == -1 ||
+	     is_ws(ts->current_char) )
+	{ ts->recovered = R_STATEMENT;
 	  break;
-	if ( ts->current_char == '.' )
-	{ if ( !next(ts) ||
-	       ts->current_char == -1 ||
-	       is_ws(ts->current_char) )
-	    break;
 	}
       }
+      if ( ( (ts->current_char == ';' && (ts->recover == R_PREDOBJLIST ||
+					  ts->recover == R_OBJLIST) ) ||
+	     (ts->current_char == ',' && ts->recover == R_OBJLIST)
+	   ) && ts->on_error == E_WARNING )
+      { ts->recovered = (ts->current_char == ';' ? R_PREDOBJLIST : R_OBJLIST);
+	break;
+      }
     }
-
-    if ( is_error && ts->on_error == E_ERROR )
-      return PL_raise_exception(ex);
-    else
-      print_warning(ex);
   }
 
-  return FALSE;
+  if ( is_error && ts->on_error == E_ERROR )
+  { return PL_raise_exception(ex);	/* raise exception */
+  } else
+  { static predicate_t print_message2;	/* or print message */
+    term_t av;
+
+    if ( !print_message2 )
+      print_message2 = PL_predicate("print_message", 2, "system");
+
+    if ( (av = PL_new_term_refs(2)) &&
+	 PL_put_atom(av+0, is_error ? ATOM_error : ATOM_warning) &&
+	 PL_put_term(av+1, ex) )
+      PL_call_predicate(NULL, PL_Q_PASS_EXCEPTION, print_message2, av);
+
+    return FALSE;
+  }
 }
 
 
+static int
+syntax_message(turtle_state *ts, const char *msg, int is_error)
+{ term_t ex1;
+
+  if ( PL_exception(0) )		/* pending earlier exception */
+    return FALSE;
+
+  ts->error_count++;
+
+  if ( !(ex1=PL_new_term_ref()) ||
+       !PL_unify_term(ex1, PL_FUNCTOR, FUNCTOR_syntax_error1,
+		              PL_CHARS, msg) )
+    return FALSE;
+
+  return print_message(ts, ex1, is_error);
+}
+
 static int
 syntax_warning(turtle_state *ts, const char *msg)
 { return syntax_message(ts, msg, FALSE);
@@ -956,23 +1008,18 @@ free_resources(turtle_state *ts)
 
 static int
 turtle_existence_error(turtle_state *ts, const char *what, term_t t)
-{ if ( ts->on_error == E_ERROR )
-  { return PL_existence_error(what, t);
-  } else
-  { term_t ex = PL_new_term_ref();
+{ term_t ex1;
 
-    if ( PL_unify_term(ex,
-		       PL_FUNCTOR, FUNCTOR_error2,
-		         PL_FUNCTOR, FUNCTOR_existence_error2,
-		           PL_CHARS, what,
-		           PL_TERM, t,
-		         PL_VARIABLE) )
-      print_warning(ex);
-
-    return FALSE;
+  if ( (ex1 = PL_new_term_ref()) &&
+       PL_unify_term(ex1,
+		     PL_FUNCTOR, FUNCTOR_existence_error2,
+		       PL_CHARS, what,
+		       PL_TERM, t) )
+  { return print_message(ts, ex1, TRUE);
   }
-}
 
+  return FALSE;
+}
 
 
 static resource *
@@ -1092,9 +1139,9 @@ make_absolute_resource(turtle_state *ts, const wchar_t *uri)
 		 *	   PARSER OBJECT	*
 		 *******************************/
 
-static void	clear_turtle_parser(turtle_state *ts);
+static int	clear_turtle_parser(turtle_state *ts);
 static int	init_base_uri(turtle_state *ts);
-static int	read_predicate_object_list(turtle_state *ts, int end);
+static int	read_predicate_object_list(turtle_state *ts, const char *end);
 static int	read_object(turtle_state *ts);
 static int	set_subject(turtle_state *ts, resource *r, resource **old);
 static int	set_predicate(turtle_state *ts, resource *r, resource **old);
@@ -1123,13 +1170,18 @@ new_turtle_parser(IOSTREAM *s)
 }
 
 
-static void
+static int
 clear_turtle_parser(turtle_state *ts)
-{ if ( ts->base_uri )	     free(ts->base_uri);
+{ int rc;
+
+  if ( ts->base_uri )	     free(ts->base_uri);
   if ( ts->empty_prefix )    free(ts->empty_prefix);
   if ( ts->bnode.buffer )    free(ts->bnode.buffer);
 
-  if ( ts->input ) PL_release_stream(ts->input);
+  if ( ts->input )
+    rc = PL_release_stream(ts->input);
+  else
+    rc = TRUE;
 
   set_subject(ts, NULL, NULL);
   set_predicate(ts, NULL, NULL);
@@ -1142,6 +1194,8 @@ clear_turtle_parser(turtle_state *ts)
   clear_hash_table(&ts->blank_node_map);
 
   memset(ts, 0, sizeof(*ts));			/* second call does nothing */
+
+  return rc;
 }
 
 
@@ -1320,11 +1374,11 @@ set_prefix(turtle_state *ts, const wchar_t *prefix, resource *r)
 
 
 static char *
-r_name(resource *r, char *buf)
+r_name(resource *r, char *buf, size_t size)
 { if ( r->type == R_RESOURCE )
-    Ssprintf(buf, "<%Ws>", r->v.r.name);
+    Ssnprintf(buf, size, "<%Ws>", r->v.r.name);
   else if ( r->type == R_BNODE )
-    Ssprintf(buf, "bnode(%ld)", (long)r->v.bnode_id);
+    Ssnprintf(buf, size, "bnode(%ld)", (long)r->v.bnode_id);
   else
     assert(0);
 
@@ -1333,17 +1387,17 @@ r_name(resource *r, char *buf)
 
 
 static char *
-o_name(object *o, char *buf)
+o_name(object *o, char *buf, size_t size)
 { if ( o->type == O_RESOURCE )
-    return r_name(o->value.r, buf);
+    return r_name(o->value.r, buf, size);
   if ( o->type == O_LITERAL )
   { if ( o->value.l.lang )
-      Ssprintf(buf, "\"%Ws\"@%Ws", o->value.l.string, o->value.l.lang);
+      Ssnprintf(buf, size, "\"%Ws\"@%Ws", o->value.l.string, o->value.l.lang);
     else if ( o->value.l.type )
-      Ssprintf(buf, "\"%Ws\"^^<%Ws>", o->value.l.string,
+      Ssnprintf(buf, size, "\"%Ws\"^^<%Ws>", o->value.l.string,
 	       o->value.l.type->v.r);
     else
-      Ssprintf(buf, "\"%Ws\"", o->value.l.string);
+      Ssnprintf(buf, size, "\"%Ws\"", o->value.l.string);
 
     return buf;
   }
@@ -1469,9 +1523,9 @@ got_triple(turtle_state *ts, resource *s, resource *p, object *o)
     char obuf[256];
 
     Sdprintf("Got %s %s %s\n",
-	     r_name(s, sbuf),
-	     r_name(p, pbuf),
-	     o_name(o, obuf));
+	     r_name(s, sbuf, sizeof(sbuf)),
+	     r_name(p, pbuf, sizeof(pbuf)),
+	     o_name(o, obuf, sizeof(obuf)));
 
     return TRUE;
   }
@@ -1888,7 +1942,7 @@ read_blank_node_property_list(turtle_state *ts)
 
   rc = ( set_anon_subject(ts, &olds) &&
 	 set_predicate(ts, NULL, &oldp) &&
-	 read_predicate_object_list(ts, ']')
+	 read_predicate_object_list(ts, "]")
        );
   set_subject(ts, olds, &bnode);
   set_predicate(ts, oldp, NULL);
@@ -2049,12 +2103,23 @@ read_iri(turtle_state *ts, int flags)
 }
 
 
+/* read_verb() reads a predicate.  If the predicate cannot be read, input
+   is advanced to the next ; or end-of-statement.
+*/
+
 static int
 read_verb(turtle_state *ts)
-{ resource *r;
+{ for(;;)
+  { resource *r;
 
-  if ( (r=read_iri(ts, IRI_VERB)) )
-    return set_predicate(ts, r, NULL);
+    RECOVER(R_PREDOBJLIST,
+	    r=read_iri(ts, IRI_VERB));
+    if ( r )
+      return set_predicate(ts, r, NULL);
+    if ( ts->recovered == R_PREDOBJLIST && next(ts) )
+      continue;
+    break;
+  }
 
   return FALSE;
 }
@@ -2167,23 +2232,27 @@ read_long_string(turtle_state *ts, int q, string_buffer *text)
 
 static int
 read_lang(turtle_state *ts, string_buffer *b)
-{ initBuf(b);
+{ int minus = 0;
+  int fminus = TRUE;
 
-  if ( is_lang_char1(ts->current_char) )
-  { addBuf(b, ts->current_char);
-  } else
-  { return syntax_error(ts, "LANGTAG expected");
-  }
+  initBuf(b);
 
   for(;;)
-  { if ( next(ts) )
-    { if ( is_lang_char(ts->current_char) )
-      { addBuf(b, ts->current_char);
-      } else
-      { addBuf(b, EOS);
-	return TRUE;
-      }
+  { if ( is_lang_char(ts->current_char, minus) )
+    { addBuf(b, ts->current_char);
+      fminus = FALSE;
+    } else if ( ts->current_char == '-' && fminus == FALSE )
+    { addBuf(b, ts->current_char);
+      minus++;
+      fminus = TRUE;
+    } else if ( fminus == FALSE )
+    { addBuf(b, EOS);
+      return TRUE;
     } else
+    { return syntax_error(ts, "LANGTAG expected");
+    }
+
+    if ( !next(ts) )
       return FALSE;
   }
 }
@@ -2331,6 +2400,7 @@ read_object(turtle_state *ts)
 	free_resource(ts, r);
 	return rc;
       }
+      return FALSE;
     }
     case '[':
     { resource *r;
@@ -2395,10 +2465,15 @@ read_object(turtle_state *ts)
       if ( read_string(ts, &text) )
       { int rc;
 
+	if ( !skip_ws(ts) )
+	{ discardBuf(&text);
+	  return FALSE;
+	}
+
 	if ( ts->current_char == '@' )
 	{ string_buffer lang;
 
-	  if ( (rc = next(ts) && read_lang(ts, &lang)) )
+	  if ( (rc = next(ts) && skip_ws(ts) && read_lang(ts, &lang)) )
 	  { rc = got_lang_triple(ts,
 				 bufSize(&text)-1, baseBuf(&text),
 				 baseBuf(&lang));
@@ -2408,7 +2483,7 @@ read_object(turtle_state *ts)
 	{ if ( next(ts) && ts->current_char == '^' )
 	  { resource *r;
 
-	    if ( next(ts) && (r=read_iri(ts, 0)) )
+	    if ( next(ts) && skip_ws(ts) && (r=read_iri(ts, 0)) )
 	    { rc = got_typed_triple(ts, bufSize(&text)-1, baseBuf(&text), r);
 	      free_resource(ts, r);
 	    } else
@@ -2453,8 +2528,11 @@ read_object(turtle_state *ts)
 static int
 read_object_list(turtle_state *ts)
 { for(;;)
-  { if ( !read_object(ts) ||
-	 !skip_ws(ts) )
+  { int  rc;
+
+    RECOVER(R_OBJLIST, rc = read_object(ts));
+
+    if ( !skip_ws(ts) )
       return FALSE;
 
     if ( ts->current_char == ',' )
@@ -2463,14 +2541,16 @@ read_object_list(turtle_state *ts)
       else
 	return FALSE;
     }
+    if ( ts->current_char == ';' )
+      return TRUE;
 
-    return TRUE;
+    return rc;
   }
 }
 
 
 static int
-read_predicate_object_list(turtle_state *ts, int end)
+read_predicate_object_list(turtle_state *ts, const char *end)
 { for(;;)
   { if ( !read_verb(ts)	||
 	 !read_object_list(ts) ||
@@ -2480,7 +2560,7 @@ read_predicate_object_list(turtle_state *ts, int end)
     if ( ts->current_char == ';' )
     { empty:
       if ( next(ts) && skip_ws(ts) )
-      { if ( ts->current_char == end )
+      { if ( ts->current_char <= 256 && strchr(end, ts->current_char) )
 	  return TRUE;
 	if ( ts->current_char == ';' )
 	  goto empty;
@@ -2645,8 +2725,20 @@ the subject.
 
 static int
 final_predicate_object_list(turtle_state *ts)
-{ return ( read_predicate_object_list(ts, '.') &&
-	   read_end_of_clause(ts) );
+{ const char *end;
+
+  if ( ts->format == D_TRIG && ts->current_graph )
+    end = "}.";
+  else
+    end = ".";
+
+  if ( !read_predicate_object_list(ts, end) )
+    return FALSE;
+
+  if ( ts->current_char == '}' && ts->format == D_TRIG && ts->current_graph )
+    return TRUE;
+
+  return read_end_of_clause(ts);
 }
 
 
@@ -2970,7 +3062,8 @@ unify_turtle_parser(term_t parser, turtle_state *ts)
 { parser_symbol *ps;
 
   if ( (ps=malloc(sizeof(*ps))) )
-  { ps->magic = PARSER_MAGIC;
+  { memset(ps, 0, sizeof(*ps));		/* ensure padding is initialized */
+    ps->magic = PARSER_MAGIC;
     ps->state = ts;
 
     return PL_unify_blob(parser, ps, sizeof(*ps), &turtle_blob);
@@ -2984,7 +3077,7 @@ static foreign_t
 create_turtle_parser(term_t parser, term_t in, term_t options)
 { IOSTREAM *stream;
 
-  if ( PL_get_stream_handle(in, &stream) )
+  if ( PL_get_stream(in, &stream, SIO_INPUT) )
   { turtle_state *ts;
 
     if ( (ts=new_turtle_parser(stream)) )
@@ -2994,7 +3087,7 @@ create_turtle_parser(term_t parser, term_t in, term_t options)
 
       while(PL_get_list_ex(opts, opt, opts))
       { atom_t name;
-	int arity;
+	size_t arity;
 
 	if ( PL_get_name_arity(opt, &name, &arity) )
 	{ if ( arity == 1 )
@@ -3104,10 +3197,7 @@ destroy_turtle_parser(term_t parser)
 { turtle_state *ts;
 
   if ( get_turtle_parser(parser, &ts) )
-  { clear_turtle_parser(ts);
-
-    return TRUE;
-  }
+    return clear_turtle_parser(ts);
 
   return FALSE;
 }
@@ -3127,7 +3217,7 @@ turtle_parse(term_t parser, term_t triples, term_t options)
 
     while(PL_get_list_ex(opts, opt, opts))
     { atom_t name;
-      int arity;
+      size_t arity;
 
       if ( PL_get_name_arity(opt, &name, &arity) )
       { if ( arity == 1 )
@@ -3323,10 +3413,100 @@ turtle_format(term_t parser, term_t format)
 
 /** turtle_pn_local(+Atom) is semidet.
 
-True if Atom is a valid Turtle   local  following the old standard. Note
-that in the current standard, any IRI is a valid pn_local string.
+True if Atom is a valid Turtle local name.
 */
 
+static const char *
+skip_plx(const char *s, const char *e)
+{ if ( s+3 < e && *s == '%' && hexd(s[1]) >= 0 && hexd(s[2]) >= 0 )
+    return s+3;
+
+  return NULL;
+}
+
+static int
+is_pn_local(const char *s, size_t len)
+{ if ( len > 0 )
+  { const char *e = &s[len];
+    int c = s[0]&0xff;
+
+    if ( pn_local_start(c) || is_local_escape(c) )
+      s++;
+    else if ( !(s=skip_plx(s, e)) )
+      return FALSE;
+
+    while(s<e)
+    { int c = s[0]&0xff;
+
+      if ( wcis_pn_chars(c) || c == ':' || is_local_escape(c) )
+      { s++;
+	continue;
+      }
+      if ( (s = skip_plx(s,e)) )
+	continue;
+      if ( c == '.' )
+      { if ( s+1 < e )
+	{ c = s[1]&0xff;
+	  if ( wcis_pn_chars(c) || c == ':' || c == '.' || c == '%' )
+	  { s++;
+	    continue;
+	  }
+	}
+      }
+      return FALSE;
+    }
+  }
+  return TRUE;
+}
+
+
+static const pl_wchar_t *
+wskip_plx(const pl_wchar_t *s, const pl_wchar_t *e)
+{ if ( s+3 < e && *s == '%' && hexd(s[1]) >= 0 && hexd(s[2]) >= 0 )
+    return s+3;
+
+  return NULL;
+}
+
+
+static int
+wis_pn_local(const pl_wchar_t *s, size_t len)
+{ if ( len > 0 )
+  { const pl_wchar_t *e = &s[len];
+    int c = s[0];
+
+    if ( pn_local_start(c) || is_local_escape(c) )
+      s++;
+    else if ( !(s=wskip_plx(s, e)) )
+      return FALSE;
+
+    while(s<e)
+    { int c = s[0];
+
+      if ( wcis_pn_chars(c) || c == ':' || is_local_escape(c) )
+      { s++;
+	continue;
+      }
+      if ( (s = wskip_plx(s,e)) )
+	continue;
+      if ( c == '.' )
+      { if ( s+1 < e )
+	{ c = s[1];
+	  if ( wcis_pn_chars(c) || c == ':' || c == '.' || c == '%' )
+	  { s++;
+	    continue;
+	  }
+	}
+      }
+      return FALSE;
+    }
+  }
+  return TRUE;
+}
+
+
+
+
 static foreign_t
 turtle_pn_local(term_t name)
 { char *s;
@@ -3334,30 +3514,45 @@ turtle_pn_local(term_t name)
   size_t len;
 
   if ( PL_get_nchars(name, &len, &s, CVT_ATOM) )
-  { const char *e = &s[len];
-
-    if ( !wcis_pn_chars_u(s[0]&0xff) )
-      return FALSE;
-    for(s++; s<e; s++)
-    { if ( !wcis_pn_chars(s[0]&0xff) )
-	return FALSE;
-    }
-    return TRUE;
-  } else if ( PL_get_wchars(name, &len, &w, CVT_ATOM|CVT_EXCEPTION) )
-  { const pl_wchar_t *e = &w[len];
-
-    if ( !wcis_pn_chars_u(w[0]) )
-      return FALSE;
-    for(w++; w<e; w++)
-    { if ( !wcis_pn_chars(w[0]) )
-	return FALSE;
-    }
-    return TRUE;
-  } else
+    return is_pn_local(s, len);
+  else if ( PL_get_wchars(name, &len, &w, CVT_ATOM|CVT_EXCEPTION) )
+    return wis_pn_local(w, len);
+  else
     return FALSE;
 }
 
 
+static foreign_t
+iri_turtle_prefix(term_t iri, term_t prefix)
+{ char *s;
+  pl_wchar_t *w;
+  size_t len;
+
+  if ( PL_get_nchars(iri, &len, &s, CVT_ATOM) )
+  { const char *e = &s[len]-1;
+
+    while(e>s && e[0] != '/' && e[0] != '#')
+      e--;
+    if ( e < &s[len] && (e[0] == '/' || e[0] == '#') )
+      e++;
+    if ( is_pn_local(e, &s[len]-e) )
+      return PL_unify_atom_nchars(prefix, e-s, s);
+  } else if ( PL_get_wchars(iri, &len, &w, CVT_ATOM|CVT_EXCEPTION) )
+  { const pl_wchar_t *e = &w[len]-1;
+
+    while(e>w && e[0] != '/' && e[0] != '#')
+      e--;
+    if ( e < &w[len] && (e[0] == '/' || e[0] == '#') )
+      e++;
+    if ( wis_pn_local(e, &w[len]-e) )
+      return PL_unify_wchars(prefix, PL_ATOM, e-w, w);
+  }
+
+  return FALSE;
+}
+
+
+
 		 /*******************************
 		 *	     WRITING		*
 		 *******************************/
@@ -3393,6 +3588,7 @@ ttl_put_character(IOSTREAM *s, int c)
 static int
 ttl_put_echaracter(IOSTREAM *s, int c)
 { int c2;
+  int rc;
 
   switch(c)
   { case '\t': c2 = 't'; break;
@@ -3402,7 +3598,8 @@ ttl_put_echaracter(IOSTREAM *s, int c)
       return ttl_put_character(s, c);
   }
 
-  Sputcode('\\', s);
+  if ( (rc=Sputcode('\\', s)) < 0 )
+    return rc;
 
   return Sputcode(c2, s);
 }
@@ -3410,12 +3607,16 @@ ttl_put_echaracter(IOSTREAM *s, int c)
 
 static int
 ttl_put_scharacter(IOSTREAM *s, int c)
-{ switch(c)
+{ int rc;
+
+  switch(c)
   { case '"':
-      Sputcode('\\', s);
+      if ( (rc=Sputcode('\\', s)) < 0 )
+	return rc;
       return Sputcode('"', s);
     case '\\':
-      Sputcode('\\', s);
+      if ( (rc=Sputcode('\\', s)) < 0 )
+	return rc;
       return Sputcode('\\', s);
     default:
       return ttl_put_echaracter(s, c);
@@ -3430,6 +3631,9 @@ write_long_q(IOSTREAM *out)
   Sputcode('"', out);
 }
 
+#define StryPutcode(c, s) \
+	do { if ( Sputcode(c,s) < 0 ) goto error; } while(0)
+
 static foreign_t
 turtle_write_quoted_string(term_t Stream, term_t Value, term_t wlong)
 { size_t len;
@@ -3468,24 +3672,25 @@ turtle_write_quoted_string(term_t Stream, term_t Value, term_t wlong)
     { write_long_q(out);
       for(s=textA; s<e; s++)
       { if ( *s == '"' && ((s+1<e && s[1] != '"') || (s+2<e && s[2] != '"')) )
-	  Sputcode('"', out);
+	  StryPutcode('"', out);
 	else if ( *s == '\n' || *s == '\r' )
-	  Sputcode(*s, out);
+	  StryPutcode(*s, out);
 	else if ( ttl_put_scharacter(out, s[0]&0xff) < 0 )
-	  break;
+	  goto error;
       }
       write_long_q(out);
     } else
-    { Sputcode('"', out);
+    { StryPutcode('"', out);
       for(s=textA; s<e; s++)
       { if ( ttl_put_scharacter(out, s[0]&0xff) < 0 )
-	  break;
+	  goto error;
       }
-      Sputcode('"', out);
+      StryPutcode('"', out);
     }
 
     return PL_release_stream(out);
-  } else if ( PL_get_wchars(Value, &len, &textW, CVT_ATOM|CVT_EXCEPTION) )
+  } else if ( PL_get_wchars(Value, &len, &textW,
+			    CVT_ATOM|CVT_STRING|CVT_EXCEPTION) )
   { const pl_wchar_t *e = &textW[len];
     const pl_wchar_t *w;
 
@@ -3509,25 +3714,26 @@ turtle_write_quoted_string(term_t Stream, term_t Value, term_t wlong)
     { write_long_q(out);
       for(w=textW; w<e; w++)
       { if ( *w == '"' && ((w+1<e && w[1] != '"') || (w+2<e && w[2] != '"')) )
-	  Sputcode('"', out);
+	  StryPutcode('"', out);
 	else if ( *w == '\n' || *w == '\r' )
-	  Sputcode(*w, out);
+	  StryPutcode(*w, out);
 	else if ( ttl_put_scharacter(out, w[0]) < 0 )
-	  break;
+	  goto error;
       }
       write_long_q(out);
     } else
-    { Sputcode('"', out);
+    { StryPutcode('"', out);
       for(w=textW; w<e; w++)
       { if ( ttl_put_scharacter(out, w[0]) < 0 )
-	  break;
+	  goto error;
       }
-      Sputcode('"', out);
+      StryPutcode('"', out);
     }
 
     return PL_release_stream(out);
   } else
-  { PL_release_stream(out);
+  { error:
+    PL_release_stream(out);
     return FALSE;
   }
 }
@@ -3535,11 +3741,22 @@ turtle_write_quoted_string(term_t Stream, term_t Value, term_t wlong)
 
 static int
 ttl_put_ucharacter(IOSTREAM *s, int c)
-{ switch(c)
-  { case '>':
+{ int rc;
+
+  switch(c)
+  { case '>':				/* not allowed in rfc3987.txt */
+    case '<':
     case '\\':
-      Sputcode('\\', s);
-      return Sputcode(c, s);
+    case '^':
+    case '|':
+    case '{':
+    case '}':
+    case '"':
+    case '`':
+    case ' ':
+      if ( (rc=Sfprintf(s, "%%%02x", c)) < 0 )
+	return rc;
+      return 0;
     default:
       return ttl_put_character(s, c);
   }
@@ -3562,25 +3779,96 @@ turtle_write_uri(term_t Stream, term_t Value)
   if ( PL_get_nchars(Value, &len, &s, CVT_ATOM|CVT_STRING) )
   { const char *e = &s[len];
 
-    Sputcode('<', out);
+    StryPutcode('<', out);
     for(; s<e; s++)
     { if ( ttl_put_ucharacter(out, s[0]&0xff) < 0 )
-	break;
+	goto error;
     }
-    Sputcode('>', out);
+    StryPutcode('>', out);
     return PL_release_stream(out);
   } else if ( PL_get_wchars(Value, &len, &w, CVT_ATOM|CVT_EXCEPTION) )
   { const pl_wchar_t *e = &w[len];
 
-    Sputcode('<', out);
+    StryPutcode('<', out);
     for(; w<e; w++)
     { if ( ttl_put_ucharacter(out, w[0]) < 0 )
-	break;
+	goto error;
     }
-    Sputcode('>', out);
+    StryPutcode('>', out);
     return PL_release_stream(out);
   } else
-  { PL_release_stream(out);
+  { error:
+    PL_release_stream(out);
+    return FALSE;
+  }
+}
+
+
+/* Local escape characters are: ~.-!$&'()*+,;=/?#@%_
+*/
+
+static inline int
+wr_is_local_escape(int c)
+{ return is_local_escape(c) && !strchr("_-%", c);
+}
+
+static foreign_t
+turtle_write_pn_local(term_t Stream, term_t Value)
+{ size_t len;
+  char *s;
+  pl_wchar_t *w;
+  IOSTREAM *out;
+
+  if ( !PL_get_stream_handle(Stream, &out) )
+    return FALSE;
+
+  if ( PL_get_nchars(Value, &len, &s, CVT_ATOM|CVT_STRING) )
+  { if ( len > 0 )
+    { const char *e = &s[len];
+      int c = s[0]&0xff;
+
+      if ( !pn_local_start(c) )
+	StryPutcode('\\', out);
+      StryPutcode(c, out);
+
+      for(s++; s<e; s++)
+      { c = s[0]&0xff;
+
+	if ( c == '.' && s+1 < e && !strchr(":.%", s[1]) )
+	{ StryPutcode(c, out);
+	} else
+	{ if ( wr_is_local_escape(c) )
+	    StryPutcode('\\', out);
+	  StryPutcode(c, out);
+	}
+      }
+    }
+    return PL_release_stream(out);
+  } else if ( PL_get_wchars(Value, &len, &w, CVT_ATOM|CVT_EXCEPTION) )
+  { if ( len > 0 )
+    { const pl_wchar_t *e = &w[len];
+      int c = w[0];
+
+      if ( !pn_local_start(c) )
+	StryPutcode('\\', out);
+      StryPutcode(c, out);
+
+      for(w++; w<e; w++)
+      { c = w[0];
+
+	if ( c == '.' && w+1 < e && !strchr(":.%", w[1]) )
+	{ StryPutcode(c, out);
+	} else
+	{ if ( wr_is_local_escape(c) )
+	    StryPutcode('\\', out);
+	  StryPutcode(c, out);
+	}
+      }
+    }
+    return PL_release_stream(out);
+  } else
+  { error:
+    PL_release_stream(out);
     return FALSE;
   }
 }
@@ -3639,7 +3927,9 @@ install_turtle(void)
   PL_register_foreign("turtle_format",	       2, turtle_format,         0);
 
   PL_register_foreign("turtle_pn_local",       1, turtle_pn_local,       0);
+  PL_register_foreign("iri_turtle_prefix",     2, iri_turtle_prefix,     0);
   PL_register_foreign("turtle_write_uri",      2, turtle_write_uri,      0);
+  PL_register_foreign("turtle_write_pn_local", 2, turtle_write_pn_local, 0);
   PL_register_foreign("turtle_write_quoted_string",
 					    3, turtle_write_quoted_string, 0);
 }
