@@ -3,22 +3,34 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2013, University of Amsterdam
-			      VU University Amsterdam
+    Copyright (c)  2011-2017, University of Amsterdam
+                              VU University Amsterdam
+    All rights reserved.
 
-    This library is free software; you can redistribute it and/or
-    modify it under the terms of the GNU Lesser General Public
-    License as published by the Free Software Foundation; either
-    version 2.1 of the License, or (at your option) any later version.
+    Redistribution and use in source and binary forms, with or without
+    modification, are permitted provided that the following conditions
+    are met:
 
-    This library is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-    Lesser General Public License for more details.
+    1. Redistributions of source code must retain the above copyright
+       notice, this list of conditions and the following disclaimer.
 
-    You should have received a copy of the GNU Lesser General Public
-    License along with this library; if not, write to the Free Software
-    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
+    2. Redistributions in binary form must reproduce the above copyright
+       notice, this list of conditions and the following disclaimer in
+       the documentation and/or other materials provided with the
+       distribution.
+
+    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
+    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
+    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
+    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
+    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
+    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
+    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
+    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
+    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
+    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
+    POSSIBILITY OF SUCH DAMAGE.
 */
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -61,12 +73,9 @@ typedef struct
 			  { FMT_ERROR("not enough arguments"); \
 			  } \
 			}
-#define FMT_ERROR(fmt)	return (void)Sunlock(fd), \
-			  PL_error(NULL, 0, NULL, ERR_FORMAT, fmt)
-#define FMT_ARG(c, a)	return (void)Sunlock(fd), \
-			       PL_error(NULL, 0, NULL, \
-					ERR_FORMAT_ARG, c, a)
-#define FMT_EXEPTION()	return (void)Sunlock(fd), FALSE
+#define FMT_ERROR(fmt)	return PL_error(NULL, 0, NULL, ERR_FORMAT, fmt)
+#define FMT_ARG(c, a)	return PL_error(NULL, 0, NULL, ERR_FORMAT_ARG, c, a)
+#define FMT_EXEPTION()	return FALSE
 
 
 static PL_locale prolog_locale =
@@ -93,7 +102,7 @@ UTF-8 format in the state's `buffer'.   The  `buffered' field represents
 the number of UTF-8 characters in the buffer.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-static int
+static WUNUSED int
 outchr(format_state *state, int chr)
 { if ( state->pending_rubber )
   { if ( chr > 0x7f )
@@ -126,7 +135,7 @@ Emit ASCII 0-terminated strings resulting from sprintf() on numeric
 arguments.  No fuzz with wide characters here.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-static int
+static WUNUSED int
 outstring(format_state *state, const char *s, size_t len)
 { const char *q;
   const char *e = &s[len];
@@ -148,7 +157,7 @@ outstring(format_state *state, const char *s, size_t len)
 }
 
 
-static int
+static WUNUSED int
 oututf8(format_state *state, const char *s, size_t len)
 { const char *e = &s[len];
 
@@ -164,13 +173,13 @@ oututf8(format_state *state, const char *s, size_t len)
 }
 
 
-static int
+static WUNUSED int
 oututf80(format_state *state, const char *s)
 { return oututf8(state, s, strlen(s));
 }
 
 
-static int
+static WUNUSED int
 outtext(format_state *state, PL_chars_t *txt)
 { switch(txt->encoding)
   { case ENC_ISO_LATIN_1:
@@ -211,8 +220,7 @@ word
 pl_format_predicate(term_t chr, term_t descr)
 { int c;
   predicate_t proc = NULL;
-  Symbol s;
-  int arity;
+  size_t arity;
 
   if ( !PL_get_char_ex(chr, &c, FALSE) )
     fail;
@@ -228,10 +236,7 @@ pl_format_predicate(term_t chr, term_t descr)
   if ( !format_predicates )
     format_predicates = newHTable(8);
 
-  if ( (s = lookupHTable(format_predicates, (void *)(intptr_t)c)) )
-    s->value = proc;
-  else
-    addHTable(format_predicates, (void *)(intptr_t)c, proc);
+  addNewHTable(format_predicates, (void *)(intptr_t)c, proc);
 
   succeed;
 }
@@ -240,7 +245,8 @@ pl_format_predicate(term_t chr, term_t descr)
 word
 pl_current_format_predicate(term_t chr, term_t descr, control_t h)
 { GET_LD
-  Symbol s = NULL;
+  intptr_t name;
+  predicate_t pred;
   TableEnum e;
   fid_t fid;
 
@@ -264,9 +270,9 @@ pl_current_format_predicate(term_t chr, term_t descr, control_t h)
   { freeTableEnum(e);
     return FALSE;
   }
-  while( (s=advanceTableEnum(e)) )
-  { if ( PL_unify_integer(chr, (intptr_t)s->name) &&
-	 PL_unify_predicate(descr, (predicate_t)s->value, 0) )
+  while( advanceTableEnum(e, (void**)&name, (void**)&pred) )
+  { if ( PL_unify_integer(chr, name) &&
+	 PL_unify_predicate(descr, pred, 0) )
     { PL_close_foreign_frame(fid);
       ForeignRedoPtr(e);
     }
@@ -306,7 +312,6 @@ format_impl(IOSTREAM *out, term_t format, term_t Args, Module m)
     PL_put_term(argv, args);
   }
 
-  startCritical;
   switch(fmt.storage)			/* format can do call-back! */
   { case PL_CHARS_RING:
     case PL_CHARS_STACK:
@@ -316,10 +321,10 @@ format_impl(IOSTREAM *out, term_t format, term_t Args, Module m)
       break;
   }
 
+  Slock(out);
   rval = do_format(out, &fmt, argc, argv, m);
+  Sunlock(out);
   PL_free_text(&fmt);
-  if ( !endCritical )
-    return FALSE;
 
   return rval;
 }
@@ -376,12 +381,9 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv, Module m)
 { GET_LD
   format_state state;			/* complete state */
   int tab_stop = 0;			/* padded tab stop */
-  Symbol s;
   unsigned int here = 0;
   int rc = TRUE;
 
-  Slock(fd);				/* buffer locally */
-
   state.out = fd;
   state.pending_rubber = 0;
   initBuffer(&state.buffer);
@@ -399,6 +401,7 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv, Module m)
     { case '~':
 	{ int arg = DEFAULT;		/* Numeric argument */
 	  int mod_colon = FALSE;	/* Used colon modifier */
+	  predicate_t proc;
 					/* Get the numeric argument */
 	  c = get_chr_from_text(fmt, ++here);
 
@@ -440,9 +443,8 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv, Module m)
 
 					/* Check for user defined format */
 	  if ( format_predicates &&
-	       (s = lookupHTable(format_predicates, (void*)((intptr_t)c))) )
-	  { predicate_t proc = (predicate_t) s->value;
-	    int arity;
+	       (proc = lookupHTable(format_predicates, (void*)((intptr_t)c))) )
+	  { size_t arity;
 	    term_t av;
 	    char buf[BUFSIZE];
 	    char *str = buf;
@@ -466,14 +468,12 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv, Module m)
 	    tellString(&str, &bufsize, ENC_UTF8);
 	    rc = PL_call_predicate(NULL, PL_Q_PASS_EXCEPTION, proc, av);
 	    toldString();
-	    if ( !rc )
-	    { if ( str != buf )
-		free(str);
-	      goto out;
-	    }
-	    oututf8(&state, str, bufsize);
+	    if ( rc )
+	      rc = oututf8(&state, str, bufsize);
 	    if ( str != buf )
 	      free(str);
+	    if ( !rc )
+	      goto out;
 
 	    here++;
 	  } else
@@ -604,7 +604,6 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv, Module m)
 		    { term_t r = PL_new_term_ref();
 
 		      PL_put_integer(r, arg);
-		      Sunlock(fd);
 		      return PL_error(NULL, 0, NULL, ERR_DOMAIN,
 				      ATOM_radix, r);
 		    }
@@ -664,11 +663,12 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv, Module m)
 		    tellString(&str, &bufsize, ENC_UTF8);
 		    rc = (*f)(argv);
 		    toldString();
-		    if ( !rc )
-		      goto out;
-		    oututf8(&state, str, bufsize);
+		    if ( rc )
+		      rc = oututf8(&state, str, bufsize);
 		    if ( str != buf )
 		      free(str);
+		    if ( !rc )
+		      goto out;
 		  } else
 		  { if ( fd->position &&
 			 fd->position->linepos == state.column )
@@ -688,11 +688,12 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv, Module m)
 		      tellString(&str, &bufsize, ENC_UTF8);
 		      rc = (*f)(argv);
 		      toldString();
-		      if ( !rc )
-		        goto out;
-		      oututf8(&state, str, bufsize);
+		      if ( rc )
+			rc = oututf8(&state, str, bufsize);
 		      if ( str != buf )
 			free(str);
+		      if ( !rc )
+			goto out;
 		    }
 		  }
 		  SHIFT;
@@ -713,11 +714,12 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv, Module m)
 		    tellString(&str, &bufsize, ENC_UTF8);
 		    rc = (int)pl_write_term(argv, argv+1);
 		    toldString();
-		    if ( !rc )
-		      goto out;
-		    oututf8(&state, str, bufsize);
+		    if ( rc )
+		      rc = oututf8(&state, str, bufsize);
 		    if ( str != buf )
 		      free(str);
+		    if ( !rc )
+		      goto out;
 		  } else
 		  { if ( fd->position &&
 			 fd->position->linepos == state.column )
@@ -736,12 +738,13 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv, Module m)
 		      str = buf;
 		      tellString(&str, &bufsize, ENC_UTF8);
 		      rc = (int)pl_write_term(argv, argv+1);
-		      if ( !rc )
-			goto out;
 		      toldString();
-		      oututf8(&state, str, bufsize);
+		      if ( rc )
+			rc = oututf8(&state, str, bufsize);
 		      if ( str != buf )
 			free(str);
+		      if ( !rc )
+			goto out;
 		    }
 		  }
 		  SHIFT;
@@ -754,25 +757,22 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv, Module m)
 		  char *str = buf;
 		  size_t bufsize = BUFSIZE;
 		  term_t ex = 0;
-		  int rval;
 
 		  if ( argc < 1 )
 		  { FMT_ERROR("not enough arguments");
 		  }
 		  tellString(&str, &bufsize, ENC_UTF8);
-		  rval = callProlog(m, argv, PL_Q_CATCH_EXCEPTION, &ex);
+		  rc = callProlog(m, argv, PL_Q_CATCH_EXCEPTION, &ex);
 		  toldString();
-		  oututf8(&state, str, bufsize);
+		  if ( rc )
+		    rc = oututf8(&state, str, bufsize);
 		  if ( str != buf )
 		    free(str);
 
-		  if ( !rval )
-		  { Sunlock(fd);
-
-		    if ( ex )
-		      return PL_raise_exception(ex);
-		    else
-		      fail;
+		  if ( !rc )
+		  { if ( ex )
+		      rc = PL_raise_exception(ex);
+		    goto out;
 		  }
 
 		  SHIFT;
@@ -840,7 +840,6 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv, Module m)
 	      default:
 	      { term_t ex = PL_new_term_ref();
 
-		Sunlock(fd);
 		PL_put_atom(ex, codeToAtom(c));
 		return PL_error("format", 2, NULL, ERR_EXISTENCE,
 				PL_new_atom("format_character"),
@@ -864,8 +863,6 @@ do_format(IOSTREAM *fd, PL_chars_t *fmt, int argc, term_t argv, Module m)
     emit_rubber(&state);
 
 out:
-  Sunlock(fd);
-
   return rc;
 }
 
@@ -903,7 +900,7 @@ emit_rubber(format_state *state)
   for(j = 0; s <= e; j++)
   { int chr;
 
-    if ( rn && r->where == j )
+    while ( rn && r->where == j )
     { size_t n;
 
       for(n=0; n<r->size; n++)
